unit CChara;

interface
uses
	SysUtils, Classes, Types, StringUnitLight, NCommon, CSongList, NCommonSong;
const
	Max_Param_Num : Integer = 4;

	Params_Line : String = '[Params]';    
	Values_Line : String = '[Values]';
	Messages_Line : String = '[Messages]';

type
	TTRPGParamCodes = (tpcNowMax,tpcNow,tpcNone);

	TBaceTRPGCharactorData = class(TSongData)
	private
		FName : String;
		FParamNames: array of string;
		FParamCodes: Array of TTRPGParamCodes;
		FParamNow : Array of Integer;
		FParamMax : Array of Integer;
		
		function GetParamCodes(Index: Integer): TTRPGParamCodes;
		function GetParamMax(Index: Integer): Integer;
		function GetParamNames(Index: Integer): string;
		function GetParamNow(Index: Integer): Integer;
		procedure SetParamCodes(Index: Integer; const Value: TTRPGParamCodes);
		procedure SetParamMax(Index:Integer;const Value: Integer);
		procedure SetParamNames(Index: Integer; const Value: string);
		procedure SetParamNow(Index:Integer;const Value: Integer);
	protected
		procedure AssignTo(Dest:TPersistent);override;

		property ParamNames[Index:Integer]:string read GetParamNames write SetParamNames;
		property ParamCodes[Index:Integer] : TTRPGParamCodes read GetParamCodes write SetParamCodes;
		property ParamNow[Index:Integer] : Integer read GetParamNow write SetParamNow;
		property ParamMax[Index:Integer] : Integer read GetParamMax write SetParamMax;
	public
		constructor Create;virtual;
	published
		property HP_Name : string index 0 read GetParamNames write SetParamNames;
		property MP_Name : string index 1 read GetParamNames write SetParamNames;
		property FP_Name : string index 2 read GetParamNames write SetParamNames;
		property FMP_Name : string index 3 read GetParamNames write SetParamNames;

		property HP_Type : TTRPGParamCodes index 0 read GetParamCodes write SetParamCodes;
		property MP_Type : TTRPGParamCodes index 1 read GetParamCodes write SetParamCodes;
		property FP_Type : TTRPGParamCodes index 2 read GetParamCodes write SetParamCodes;
		property FMP_Type : TTRPGParamCodes index 3 read GetParamCodes write SetParamCodes;

		property HP_Now : Integer index 0 read GetParamNow write SetParamNow;
		property MP_Now : Integer index 1 read GetParamNow write SetParamNow;
		property FP_Now : Integer index 2 read GetParamNow write SetParamNow;
		property FMP_Now : Integer index 3 read GetParamNow write SetParamNow;

		property HP_Max : Integer index 0 read GetParamMax write SetParamMax;
		property MP_Max : Integer index 1 read GetParamMax write SetParamMax;
		property FP_Max : Integer index 2 read GetParamMax write SetParamMax;
		property FMP_Max : Integer index 3 read GetParamMax write SetParamMax;
		 
		property Name : String read FName write FName;
  end;

	TTRPGCharactorData = class(TBaceTRPGCharactorData)
	private
		FMessage : TStringList;
		FValues : TStringList;

		procedure SetMessage(const value : TStringList);
		procedure SetValues(const value : TStringList);
	protected
		procedure AssignTo(Dest:TPersistent);override;

		procedure LoadFromStream(Stream:TStream);override;
		procedure SaveToStream(Stream:TStream);override;
	public
		constructor Create;override;
		procedure LoadFromFile(const FileName:String);
		procedure SaveToFile(const FileName:String);
		destructor Destroy; override;
	published
		property Message : TStringList read FMessage write SetMessage;
		property Values : TStringList read FValues write SetValues;
	end;

	TTRPGVariable = class(TSongData)
	private
		FName:String;
		FValue:String;
		FTags:TStringList;

		procedure SetTags(const value : TStrings);
		function GetLine: String;
		procedure SetLine(Value: String);
		function GetTags: TStrings;
		function GetTagList: String;
		function GetSortTagMode: Boolean;
		procedure SetSortTagMode(const Value: Boolean);
	protected
		procedure AssignTo(Dest:TPersistent);override;
	public
		procedure LoadFromStream(Stream:TStream);override;
		procedure SaveToStream(Stream:TStream);override;
		destructor Destroy; override;
	published
		property Name : String read FName write FName;
		property Value : String read FValue write FValue;
		property Tags : TStrings read GetTags write SetTags;
		property TagList : String read GetTagList;
		property Line: String read GetLine write SetLine;
		procedure SortTags;
		property SortTagMode: Boolean read GetSortTagMode write SetSortTagMode;
	end;

	TTRPGVariableList = Class(TSongDataArray)
	private
		procedure SetItems(ID:Integer;const val:TTRPGVariable);
		function GetItems(ID:Integer):TTRPGVariable;
	protected
		procedure AssignCommonData(Dest:TPersistent);override;
		function AddID(const ID:Integer):TSongData;override;
	public
		function Add:TTRPGVariable;
		procedure Insert(ID:Integer;val:TTRPGVariable);
		property Items[ID:Integer] : TTRPGVariable read GetItems write
SetItems;default;
		function IndexOf(const S : String): Integer;
		procedure Sort;
	End;

	TTRPGResLine = class(TSongData)
	private
		FText:String;
		FTags:TStringList;

		procedure SetTags(const value : TStrings);
		function GetTags: TStrings;
	protected
		procedure AssignTo(Dest:TPersistent);override;
	public
		procedure LoadFromStream(Stream:TStream);override;
		procedure SaveToStream(Stream:TStream);override;
		destructor Destroy;override;
	published
		property Text : String read FText write FText;
		property Tags : TStrings read GetTags write SetTags;
	end;

	TTRPGResLineList = Class(TSongDataArray)
	private
		procedure SetItems(ID:Integer;const val:TTRPGResLine);
		function GetItems(ID:Integer):TTRPGResLine;
	protected
		procedure AssignCommonData(Dest:TPersistent);override;
		function AddID(const ID:Integer):TSongData;override;
	public
		function Add:TTRPGResLine;
		procedure Insert(ID:Integer;val:TTRPGResLine);
		property Items[ID:Integer] : TTRPGResLine read GetItems write SetItems;default;
		procedure SetTagValues(const Values: TTRPGVariableList);
	End;

	TTRPGBuff = class(TSongData)
	private
		FName:String;
		FEffectDices : Integer;
		FEffectVal : Integer;
		FTags:  TStringList;
		FLeft : Integer;
		FChecked : Boolean;

		procedure SetTags(const value : TStrings);
		function GetTags: TStrings;
	protected
		procedure AssignTo(Dest:TPersistent);override;
	public
		Constructor Create;
		procedure LoadFromStream(Stream:TStream);override;
		procedure SaveToStream(Stream:TStream);override;
		destructor Destroy; override;
	published
		property Name : String read FName write FName;
		property EffectDices : Integer read FEffectDices write FEffectDices;
		property EffectVal : Integer read FEffectVal write FEffectVal;
		property Tags : TStrings read GetTags write SetTags;
		property Left : Integer read FLeft write FLeft;
		property Checked : Boolean read FChecked write FChecked;
	end;

	TTRPGBuffList = Class(TSongDataArray)
	private
		procedure SetItems(ID:Integer;const val:TTRPGBuff);
		function GetItems(ID:Integer):TTRPGBuff;
	protected
		procedure AssignCommonData(Dest:TPersistent);override;
		function AddID(const ID:Integer):TSongData;override;
	public
		function Add:TTRPGBuff;
		procedure Insert(ID:Integer;val:TTRPGBuff);
		property Items[ID:Integer] : TTRPGBuff read GetItems write SetItems;default;
		procedure CountDown;
		procedure FindBuff(const Tags:TStrings;const Res: TTRPGBuff);
	End;

	TDecodedTRPGCharactorData = class(TBaceTRPGCharactorData)
	private
		FMessage : TTRPGResLineList;
		FValues : TTRPGVariableList;
		FBuffs : TTRPGBuffList;

		procedure SetMessage(const value : TTRPGResLineList);
		procedure SetValues(const value : TTRPGVariableList);
    procedure SetBuffs(const Value: TTRPGBuffList);
	protected
		procedure AssignTo(Dest:TPersistent);override;

		procedure LoadFromStream(Stream:TStream);override;
		procedure SaveToStream(Stream:TStream);override;
	public
		procedure Assign(Source:TPersistent);override;
		constructor Create; override;
		destructor Destroy; override;
	published
		property Message : TTRPGResLineList read FMessage write SetMessage;
		property Values : TTRPGVariableList read FValues write SetValues;
		property Buffs : TTRPGBuffList read FBuffs write SetBuffs;
	end;

	TTRPGCharactorDataList = Class(TSongDataArray)
	private
		procedure SetItems(ID:Integer;const val:TDecodedTRPGCharactorData);
		function GetItems(ID:Integer):TDecodedTRPGCharactorData;
	protected
		procedure AssignCommonData(Dest:TPersistent);override;
		function AddID(const ID:Integer):TSongData;override;
	public
		function Add:TDecodedTRPGCharactorData;
		procedure Insert(ID:Integer;val:TDecodedTRPGCharactorData);
		property Items[ID:Integer] : TDecodedTRPGCharactorData read GetItems 
write SetItems;default;
	End;

implementation

{ TBaceTRPGCharactorData }

procedure TBaceTRPGCharactorData.AssignTo(Dest: TPersistent);
var LC : Integer;
begin
	if Dest is TBaceTRPGCharactorData then begin
		TBaceTRPGCharactorData(Dest).Name := FName;
		for LC := 0 to Max_Param_Num - 1 do begin
			TBaceTRPGCharactorData(Dest).ParamNames[LC] := FParamNames[LC];
			TBaceTRPGCharactorData(Dest).ParamCodes[LC] := FParamCodes[LC];
			TBaceTRPGCharactorData(Dest).ParamNow[LC] := FParamNow[LC];
			TBaceTRPGCharactorData(Dest).ParamMax[LC] := FParamMax[LC];
		end;
	end else inherited;
end;

constructor TBaceTRPGCharactorData.Create;
begin
	inherited;
             
	SetLength(FParamNames,Max_Param_Num);
	SetLength(FParamCodes,Max_Param_Num);
	SetLength(FParamNow,Max_Param_Num);
	SetLength(FParamMax,Max_Param_Num);
end;

function TBaceTRPGCharactorData.GetParamCodes(Index: Integer): TTRPGParamCodes;
begin
	Result := FParamCodes[Index];
end;

function TBaceTRPGCharactorData.GetParamMax(Index: Integer): Integer;
begin
	Result := FParamMax[Index];
end;

function TBaceTRPGCharactorData.GetParamNames(Index: Integer): string;
begin
	Result := FParamNames[Index];
end;

function TBaceTRPGCharactorData.GetParamNow(Index: Integer): Integer;
begin
	Result := FParamNow[Index];
end;

procedure TBaceTRPGCharactorData.SetParamCodes(Index: Integer;
	const Value: TTRPGParamCodes);
begin
	FParamCodes[Index] := Value;
end;

procedure TBaceTRPGCharactorData.SetParamMax(Index: Integer;
	const Value: Integer);
begin
	FParamMax[Index] := Value;
end;

procedure TBaceTRPGCharactorData.SetParamNames(Index: Integer;
	const Value: string);
begin
	FParamNames[Index] := Value;
end;

procedure TBaceTRPGCharactorData.SetParamNow(Index: Integer;
	const Value: Integer);
begin
	FParamNow[Index] := Value;
end;

{TTRPGCharactorData Funx.}
procedure TTRPGCharactorData.AssignTo(Dest: TPersistent);
begin
	if Dest is TTRPGCharactorData then begin
		inherited;
		TTRPGCharactorData(Dest).Message.Text := FMessage.Text;
		TTRPGCharactorData(Dest).Values.Text := FValues.Text;
	end else inherited;
end;

procedure TTRPGCharactorData.SetMessage(const value : TStringList);
begin
	FMessage.Assign(value);
end;


procedure TTRPGCharactorData.SetValues(const value : TStringList);
begin
	FValues.Assign(value);
end;

constructor TTRPGCharactorData.Create;
begin
	inherited;

	FMessage := TStringList.Create;
	FValues := TStringList.Create;
end;

destructor TTRPGCharactorData.Destroy;
begin
	FMessage.Free;
	FValues.Free;
	inherited;
end;

procedure TTRPGCharactorData.LoadFromFile(const FileName: String);
var
	SL : TStringList;
	LC: Integer;
	procedure GetParams(var LineNum: Integer);
	var
		S,SubS : String;
		LC,I : Integer;
	begin
		Inc(LineNum);

		for LC := 0 to Max_Param_Num - 1 do begin
			S := SL[LineNum];
			if (S = Values_Line) or (S = Messages_Line) then Exit;

			I := AnsiPosBackward('=',S);
			SubS := Copy(S,I + 1,MaxInt);
			Delete(S,I,MaxInt);

			ParamNames[LC] := S;
			if InStr('/',SubS) then begin
				ParamCodes[LC] := tpcNowMax;
				TryStrToInt(trim(ExtractWordDem(SubS,'/')),I);
				ParamNow[LC] := I;

				TryStrToInt(trim(SubS),I);
				ParamMax[LC] := I;
			end else if TryStrToInt(trim(SubS),I) then begin
				ParamCodes[LC] := tpcNow;
				ParamNow[LC] := I;
				ParamMax[LC] := 0;
			end else begin
				ParamCodes[LC] := tpcNone;
				ParamNow[LC] := 0;
				ParamMax[LC] := 0;
			end;
			Inc(LineNum);
		end;
	end;          
	procedure GetDatas(const Lines: TStringList;var LineNum: Integer);
	var LC : Integer;
	begin
		Lines.Clear;
		Inc(LineNum);
		while LineNum < SL.Count do begin
			if (SL[LineNum] = Values_Line) or (SL[LineNum] = Params_Line) or
				(SL[LineNum] = Messages_Line) then
					break;

			Lines.Add(SL[LineNum]);
			Inc(LineNum);
		end;
		for LC := Lines.Count - 1 downto 0 do
			if trim(Lines[LC]) = '' then Lines.Delete(LC) else break;
	end;
begin
	SL := TStringList.Create;
	SL.LoadFromFile(FileName);

	FName := SL[0];
	LC := 1;
	while LC < SL.Count do begin
		if SL[LC] = Messages_Line then GetDatas(FMessage,LC)
		else if SL[LC] = Params_Line then GetParams(LC)
		else if SL[LC] = Values_Line then GetDatas(FValues,LC)
		else if trim(SL[LC]) = '' then Inc(LC)
		else break;
	end;
end;

procedure TTRPGCharactorData.LoadFromStream(Stream: TStream);
begin
end;

procedure TTRPGCharactorData.SaveToFile(const FileName: String);
var
	SL : TStringList;
	LC: Integer;
begin
	SL := TStringList.Create;
	SL.Add(FName);

	SL.Add(Params_Line);
	for LC := 0 to Max_Param_Num - 1 do begin
		case ParamCodes[LC] of
			tpcNowMax : SL.Add(ParamNames[LC] + '=' +
				IntToStr(ParamNow[LC]) + '/' + IntToStr(ParamMax[LC]));
			tpcNow : SL.Add(ParamNames[LC] + '=' + IntToStr(ParamNow[LC]));
			tpcNone : SL.Add(ParamNames[LC] + '=');
		end;
	end;

	SL.Add(Values_Line);
	SL.AddStrings(FValues);
	SL.Add(Messages_Line);
	SL.AddStrings(FMessage);

	SL.SaveToFile(FileName);
	SL.Free;
end;

procedure TTRPGCharactorData.SaveToStream(Stream: TStream);
begin
end;
     
{TTRPGVariable Funx.}
procedure TTRPGVariable.AssignTo(Dest: TPersistent);
begin
	if Dest is TTRPGVariable then begin
		TTRPGVariable(Dest).Name := FName;
		TTRPGVariable(Dest).Value := FValue;
		TTRPGVariable(Dest).Tags.Text := FTags.Text;
	end else inherited;
end;

procedure TTRPGVariable.SetLine(Value: String);
var
	LC : Integer;
begin
	if not Assigned(FTags) then FTags := TStringList.Create;

	FName := trimJP(ExtractWordDem(Value,'='));
	Value := TrimJP(Value);
	Tags.Text := '';

	while EndsStr('}',Value) do begin
		LC := AnsiPosBackward('{',Value);
		Delete(Value,Length(Value),1);
		Tags.Insert(0,Copy(Value,LC + 1,MaxInt));
		Delete(Value,LC,MaxInt);
		Value := TrimRightJP(Value);
	end;

	if (Tags.Count > 0) and
		((Length(Value) = 0) or (Value[Length(Value)] in ['+','-','*','/','^','%'])) then begin
			FValue := Value + '{' + Tags[0] + '}';
			Tags.Delete(0);
	end else begin
		FValue := Value;
	end;
	  
	SortTags;
end;

procedure TTRPGVariable.SetSortTagMode(const Value: Boolean);
begin
	if not Assigned(FTags) then FTags := TStringList.Create;
                                 
	if Value and (not FTags.Sorted) then SortTags;
	FTags.Sorted := Value;
	FTags.Duplicates := dupIgnore;
end;

procedure TTRPGVariable.SetTags(const value : TStrings);
begin
	if not Assigned(FTags) then FTags := TStringList.Create;
	FTags.Assign(value);
end;

procedure TTRPGVariable.SortTags;
var LC : Integer;
begin
	FTags.Sort;

	for LC := FTags.Count - 1 downto 1 do begin
		if (FTags[LC] = FTags[LC - 1]) or (FTags[LC] = '') then
			FTags.Delete(LC);
	end;
end;

destructor TTRPGVariable.Destroy;
begin
	FTags.Free;
	inherited;
end;

function TTRPGVariable.GetLine: String;
var
	S : String;
	LC: Integer;
begin          
	if not Assigned(FTags) then FTags := TStringList.Create;
	S := '';
	for LC := 0 to FTags.Count - 1 do begin
		if trim(FTags[LC]) <> '' then
			S := S + '{' + FTags[LC] + '}';
	end;

	Result := FName + ' = ' + FValue + ' ' + S;
end;

function TTRPGVariable.GetSortTagMode: Boolean;
begin                             
	if not Assigned(FTags) then FTags := TStringList.Create;
	Result := FTags.Sorted;
end;

function TTRPGVariable.GetTagList: String;
begin
	if not Assigned(FTags) then FTags := TStringList.Create;

	Result := ReplaceStr(FTags.Text,#13#10,',');
	if EndsStr(',',Result) then Delete(Result,Length(Result),1);
end;

function TTRPGVariable.GetTags: TStrings;
begin
	if not Assigned(FTags) then FTags := TStringList.Create; 
	Result := FTags;
end;

procedure TTRPGVariable.LoadFromStream(Stream: TStream);
begin
	FName := GetShortString(Stream);
	FValue := GetShortString(Stream);
	FTags.LoadFromStream(Stream);
end;

procedure TTRPGVariable.SaveToStream(Stream: TStream);
begin
	SetShortString(FName,Stream);
	SetShortString(FValue,Stream);
	FTags.SaveToStream(Stream);
end;

{TTRPGVariableList Funx.}
function TTRPGVariableList.GetItems(ID: Integer):TTRPGVariable;
begin
	Result := TTRPGVariable(inherited Items[ID]);
end;

procedure TTRPGVariableList.SetItems(ID: Integer; const val: TTRPGVariable);
begin
	inherited Items[ID] := val;
end;


function SortFunc(Item1,Item2:Pointer) : Integer;
begin
	Result := CompareStr(TTRPGVariable(Item1).Name,TTRPGVariable(Item2).Name);
end;

procedure TTRPGVariableList.Sort;
begin
	inherited Sort(SortFunc);
end;

procedure TTRPGVariableList.AssignCommonData(Dest: TPersistent);
begin

end;

function TTRPGVariableList.AddID(const ID: Integer):TSongData;
begin
	FItems[ID] := TTRPGVariable.Create;
	Result := FItems[ID];
end;

function TTRPGVariableList.Add;
begin
	Result := TTRPGVariable(inherited Add);
end;

function TTRPGVariableList.IndexOf(const S: String): Integer;
var
	L, H, I, C: Integer;
begin
	L := 0;
	H := Count - 1;
	while L <= H do begin
		I := (L + H) shr 1;
		C := CompareStr(Items[I].Name, S);
		if C < 0 then L := I + 1 else begin
			H := I - 1;
			if C = 0 then begin
				Result := I;
				Exit;
			end;
		end;
	end;
	Result := - 1;
end;

procedure TTRPGVariableList.Insert(ID: Integer; val: TTRPGVariable);
begin
	inherited insert(id,val);
end;
          

{TTRPGResLine Funx.}
procedure TTRPGResLine.AssignTo(Dest: TPersistent);
begin
	if Dest is TTRPGResLine then begin
		TTRPGResLine(Dest).Text := FText;
		TTRPGResLine(Dest).Tags := FTags;
	end else inherited;
end;

procedure TTRPGResLine.SetTags(const value : TStrings);
begin                    
	if not Assigned(FTags) then begin
		FTags := TStringList.Create;
		FTags.Sorted := True;
		FTags.Duplicates := dupIgnore;
	end;

	FTags.Assign(value);
end;

destructor TTRPGResLine.Destroy;
begin
	FTags.Free;
	inherited;
end;

function TTRPGResLine.GetTags: TStrings;
begin
	if not Assigned(FTags) then begin
		FTags := TStringList.Create;
		FTags.Sorted := True;
		FTags.Duplicates := dupIgnore;
	end;
	Result := FTags;
end;

procedure TTRPGResLine.LoadFromStream(Stream: TStream);
begin
	FText := GetShortString(Stream);
	FTags.LoadFromStream(Stream);
end;

procedure TTRPGResLine.SaveToStream(Stream: TStream);
begin
	SetShortString(FText,Stream);
	FTags.SaveToStream(Stream);
end;

{TTRPGResLineList Funx.}
function TTRPGResLineList.GetItems(ID: Integer):TTRPGResLine;
begin
	Result := TTRPGResLine(inherited Items[ID]);
end;

procedure TTRPGResLineList.SetItems(ID: Integer; const val: TTRPGResLine);
begin
	inherited Items[ID] := val;
end;

procedure TTRPGResLineList.SetTagValues(const Values: TTRPGVariableList);
var
	LC, I: Integer;
	S, SubS, Tag : String;
begin
	for LC := 0 to Count - 1 do begin
		S := Items[LC].Text;

		while InStr('{',S) do begin
			SubS := ExtractWordDem(S,'{');
			Tag := ExtractWordDem(S,'}');
			Items[LC].Tags.Add(Tag);
			I := Values.IndexOf(Tag);
			if I >= 0 then begin
				Tag := Values[I].Value;
				Items[LC].Tags.AddStrings(Values[I].Tags);
			end else Tag := '';

			S := SubS + Tag + S;
		end;
		Items[LC].Text := S;
	end;
end;

procedure TTRPGResLineList.AssignCommonData(Dest: TPersistent);
begin

end;

function TTRPGResLineList.AddID(const ID: Integer):TSongData;
begin
	FItems[ID] := TTRPGResLine.Create;
	Result := FItems[ID];
end;

function TTRPGResLineList.Add;
begin
	Result := TTRPGResLine(inherited Add);
end;

procedure TTRPGResLineList.Insert(ID: Integer; val: TTRPGResLine);
begin
	inherited insert(id,val);
end;


{ TDecodedTRPGCharactorData }

procedure TDecodedTRPGCharactorData.Assign(Source: TPersistent);
var
	D : TTRPGCharactorData;
	LC,REPC : Integer;
	
	function ValReproduce(const ID : Integer): String;
	var S,SubS,Tag : String;
	begin
		if ID = - 1 then begin
			Result := '';
			Exit;
		end;
		Inc(Repc);
		if Repc > D.Values.Count + 5 then begin
			Result := '';
			Exit;
		end;

		S := FValues[ID].Value;
		FValues[ID].SortTagMode := True;

		while InStr('{',S) do begin
			SubS := ExtractWordDem(S,'{');
			Tag := ExtractWordDem(S,'}');
			FValues[ID].Tags.Add(Tag);
			Tag := ValReproduce(FValues.IndexOf(Tag));
			S := SubS + Tag + S;
		end;
		Result := S;
		FValues[ID].Value := S;
	end;
begin
	REPC := 0;

	if Source is TTRPGCharactorData then begin
		Inherited; //TBaceTRPGCharactorData
		D := TTRPGCharactorData(Source);
		FMessage.Clear;
		FValues.Clear;
		FBuffs.Clear;

		for LC := 0 to D.Values.Count - 1 do begin
			if D.Values[LC] <> '' then FValues.Add.Line := D.Values[LC];
		end;
		FValues.Sort;
		for LC := 0 to FValues.Count - 1 do ValReproduce(LC);
                
		for LC := 0 to D.Message.Count - 1 do begin
			if D.Message[LC] <> '' then begin
				with FMessage.Add do begin
					Text := D.Message[LC];
				end;
			end;
		end;

		FMessage.SetTagValues(FValues);
	end else inherited;
end;

procedure TDecodedTRPGCharactorData.AssignTo(Dest: TPersistent);
begin
	if Dest is TDecodedTRPGCharactorData then begin
		Inherited;

		TDecodedTRPGCharactorData(Dest).Message := FMessage;
		TDecodedTRPGCharactorData(Dest).Values := FValues;
		TDecodedTRPGCharactorData(Dest).Buffs := FBuffs;
	end;
end;

constructor TDecodedTRPGCharactorData.Create;
begin
	inherited;
	FMessage := TTRPGResLineList.Create;
	FValues := TTRPGVariableList.Create;
	FBuffs := TTRPGBuffList.Create;
end;

destructor TDecodedTRPGCharactorData.Destroy;
begin
	FMessage.Free;
	FValues.Free;
	FBuffs.Free;
	inherited;
end;

procedure TDecodedTRPGCharactorData.LoadFromStream(Stream: TStream);
begin
	inherited;
end;

procedure TDecodedTRPGCharactorData.SaveToStream(Stream: TStream);
begin
	inherited;
end;

procedure TDecodedTRPGCharactorData.SetBuffs(const Value: TTRPGBuffList);
begin
	if Assigned(Value) then FBuffs.Assign(Value) else FBuffs.Clear;
end;

procedure TDecodedTRPGCharactorData.SetMessage(const value: TTRPGResLineList);
begin
	if Assigned(Value) then FMessage.Assign(Value) else FMessage.Clear;
end;

procedure TDecodedTRPGCharactorData.SetValues(const value: TTRPGVariableList);
begin
	if Assigned(Value) then FValues.Assign(Value) else FValues.Clear;
end;

{TTRPGCharactorDataList Funx.}
function TTRPGCharactorDataList.GetItems(ID: Integer):TDecodedTRPGCharactorData;
begin
	Result := TDecodedTRPGCharactorData(inherited Items[ID]);
end;

procedure TTRPGCharactorDataList.SetItems(ID: Integer; const val:
TDecodedTRPGCharactorData);
begin
	inherited Items[ID] := val;
end;

procedure TTRPGCharactorDataList.AssignCommonData(Dest: TPersistent);
begin

end;

function TTRPGCharactorDataList.AddID(const ID: Integer):TSongData;
begin
	FItems[ID] := TDecodedTRPGCharactorData.Create;
	Result := FItems[ID];
end;

function TTRPGCharactorDataList.Add;
begin
	Result := TDecodedTRPGCharactorData(inherited Add);
end;

procedure TTRPGCharactorDataList.Insert(ID: Integer; val: TDecodedTRPGCharactorData);
begin
	inherited insert(id,val);
end;

{TTRPGBuff Funx.}
procedure TTRPGBuff.AssignTo(Dest: TPersistent);
begin
	if Dest is TTRPGBuff then begin
		TTRPGBuff(Dest).Name := FName;
		TTRPGBuff(Dest).EffectDices := FEffectDices;
		TTRPGBuff(Dest).EffectVal := FEffectVal;
		TTRPGBuff(Dest).Tags := FTags;
		TTRPGBuff(Dest).Left := FLeft;      
		TTRPGBuff(Dest).Checked := FChecked;
	end else inherited;
end;

procedure TTRPGBuff.SetTags(const value : TStrings);
begin
	FTags.Assign(value);
end;

constructor TTRPGBuff.Create;
begin
	FTags := TStringList.Create;
	FTags.Sorted := True;
	FTags.Duplicates := dupIgnore;
	FChecked := True;
end;

destructor TTRPGBuff.Destroy;
begin
	FTags.Free;
	inherited;
end;

function TTRPGBuff.GetTags: TStrings;
begin
	Result := FTags;
end;

procedure TTRPGBuff.LoadFromStream(Stream: TStream);
begin
	FName := GetShortString(Stream);
	Stream.Read(FChecked,Sizeof(Checked));
	Stream.Read(FEffectDices,Sizeof(EffectDices));
	Stream.Read(FEffectVal,Sizeof(EffectVal));
	Stream.Read(FLeft,Sizeof(Left));
	FTags.LoadFromStream(Stream);
end;

procedure TTRPGBuff.SaveToStream(Stream: TStream);
begin
	SetShortString(FName,Stream);       
	Stream.Write(FChecked,Sizeof(Checked));
	Stream.Write(FEffectDices,Sizeof(EffectDices));
	Stream.Write(FEffectVal,Sizeof(EffectVal));
	Stream.Write(FLeft,Sizeof(Left));
	FTags.SaveToStream(Stream);
end;

{TTRPGBuffList Funx.}

function TTRPGBuffList.GetItems(ID: Integer):TTRPGBuff;
begin
	Result := TTRPGBuff(inherited Items[ID]);
end;

procedure TTRPGBuffList.SetItems(ID: Integer; const val: TTRPGBuff);
begin
	inherited Items[ID] := val;
end;

procedure TTRPGBuffList.AssignCommonData(Dest: TPersistent);
begin

end;

procedure TTRPGBuffList.CountDown;
var
  LC: Integer;
begin
	for LC := Count - 1 downto 0 do begin
		if Items[LC].Left > 0 then begin
			Items[LC].Left := Items[LC].Left - 1;
			if Items[LC].Left = 0 then Delete(LC);
		end;
	end;
end;

procedure TTRPGBuffList.FindBuff(const Tags: TStrings; const Res: TTRPGBuff);
var
  LC: Integer;
	MC: Integer;
begin
	Res.Name := '';
	Res.EffectDices := 0;
	Res.EffectVal := 0;
	Res.Tags.Text := '';
	Res.Left := 0;

	for LC := 0 to Count - 1 do begin
		if not Items[LC].Checked then Continue;
		for MC := 0 to Items[LC].Tags.Count - 1 do begin
			if Tags.IndexOf(Items[LC].Tags[MC]) >= 0 then begin
				Res.EffectDices := Res.EffectDices + Items[LC].EffectDices;
				Res.EffectVal := Res.EffectVal + Items[LC].EffectVal;
				if Items[LC].Name <> '' then
					Res.Name := Res.Name + ',' + Items[LC].Name;
				break;
			end;
		end;
	end;
	if Res.Name <> '' then Res.Name := DeleteStr(Res.Name,1,1);
end;

function TTRPGBuffList.AddID(const ID: Integer):TSongData;
begin
	FItems[ID] := TTRPGBuff.Create;
	Result := FItems[ID];
end;

function TTRPGBuffList.Add;
begin
	Result := TTRPGBuff(inherited Add);
end;

procedure TTRPGBuffList.Insert(ID: Integer; val: TTRPGBuff);
begin
	inherited insert(id,val);
end;

end.
