unit CSongList;
interface
uses
	Classes,SysUtils,StringUnitLight,NCommon,NCommonSong{,NCommonSRC},RTLConsts;

type
	TSongData = Class(TPersistent)
	private
	protected
	public
		procedure LoadFromStream(Stream:TStream);virtual;abstract;
		procedure SaveToStream(Stream:TStream);virtual;abstract;
	published
	End;

	TSongDataArray = Class(TSongData)
	private
		FCount : Integer;

		function GetCapacity:Integer;
		procedure SetCapacity(const val:Integer);
	protected
		FItems : Array of TSongData;

		function GetItems(ID:Integer):TSongData;
		procedure SetItems(ID:Integer;val:TSongData);

		procedure AssignTo(Dest:TPersistent);override;
		procedure AssignCommonData(Dest:TPersistent);virtual;

		function AddID(const ID:Integer):TSongData;virtual;abstract; 
		procedure Sort(Compare: TListSortCompare);
	public
		Destructor Destroy;override;
		property Items[ID:Integer]:TSongData read GetItems write SetItems;default;

		function Add:TSongData;
		procedure Delete(ID:Integer);
		procedure Insert(ID:Integer;val:TSongData);
		procedure Clear;

		procedure LoadFromStream(Stream : TStream);override;
		procedure SaveToStream(Stream : TStream);override;      
		procedure LoadFromFile(const FileName:String);
		procedure SaveToFile(const FileName:String);
	published
		property Count : Integer read FCount;
		property Capacity: Integer read GetCapacity write SetCapacity;
	End;

	TSongDataLinker = Class(TSongData)
	private
		FCount : Integer;

		function GetCapacity:Integer;
		procedure SetCapacity(const val:Integer);
	protected
		FItems : Array of TSongData;

		function GetItems(ID:Integer):TSongData;
		procedure SetItems(ID:Integer;val:TSongData);

		procedure AssignTo(Dest:TPersistent);override;
		procedure AssignCommonData(Dest:TPersistent);virtual;

		procedure Sort(Compare: TListSortCompare);
	public
		Destructor Destroy;override;
		property Items[ID:Integer]:TSongData read GetItems write SetItems;default;
		function IndexOf(Item:TSongData):Integer;

		function Add(Item:TSongData): Integer;
		procedure Delete(ID:Integer);
		procedure Insert(ID:Integer;val:TSongData);
		procedure Clear;

		procedure LoadFromStream(Stream : TStream);override;
		procedure SaveToStream(Stream : TStream);override;      
		procedure LoadFromFile(const FileName:String);
		procedure SaveToFile(const FileName:String);
	published
		property Count : Integer read FCount;
		property Capacity: Integer read GetCapacity write SetCapacity;
	End;

	TSongDataList = Class(TSongData)
	private
		function GetCount:Integer;  
		procedure SetItems(ID:Integer;const val:TSongData);
		function GetItems(ID:Integer):TSongData;
	protected
		FItems:Array of TSongData;
		function AddID(const ID:Integer):TSongData;virtual;abstract;

		procedure AssignTo(Dest:TPersistent);override;
		procedure Sort(Compare: TListSortCompare);
	public
		procedure LoadFromFile(const FileName:String);
		procedure SaveToFile(const FileName:String);
		procedure LoadFromStream(Stream : TStream);override;
		procedure SaveToStream(Stream : TStream);override;

		function Add(out ID : integer):TSongData;overload;
		function Add():TSongData;overload;

		Destructor Destroy;override;
		procedure Clear;
		procedure Delete(ID:Integer);
		property Items[ID:Integer]:TSongData read GetItems write SetItems;default;
	published
		property Count:Integer read GetCount;
	end;

implementation

{TSongDataArray Funx.}

Destructor TSongDataArray.Destroy;
begin
	Clear;
	inherited;
end;


procedure TSongDataArray.AssignCommonData(Dest: TPersistent);
var
  SourceName: string;
begin
	if Dest <> nil then
		SourceName := Dest.ClassName else SourceName := 'nil';
	raise EConvertError.CreateResFmt(@SAssignError, [ClassName, SourceName]);
end;

procedure TSongDataArray.AssignTo(Dest: TPersistent);
var List_CNT: Integer;
begin
	if Dest is TSongDataArray then begin
		TSongDataArray(Dest).Clear;

		TSongDataArray(Dest).Capacity := GetCapacity;
		AssignCommonData(Dest);

		for List_CNT := 0 to FCount - 1 do begin
			TSongDataArray(Dest).Add.Assign(FItems[List_CNT]);
		end;
	end else inherited;
end;

function TSongDataArray.GetItems(ID: Integer):TSongData;
begin
	if (ID < 0) or (ID >= FCount) then
		Result := NIL
	else Result := FItems[ID];
end;

procedure TSongDataArray.SetItems(ID: Integer; val: TSongData);
begin
	FItems[ID].Assign(val);
end;

function TSongDataArray.GetCapacity;
begin
	if Assigned(FItems) then Result := Length(FItems)
	else Result := 0;
end;

procedure TSongDataArray.SetCapacity(const val: Integer);
var
	LC: Integer;
begin
	if Val < GetCapacity then begin
		for LC := Val to FCount - 1 do FItems[LC].Free;
		SetLength(FItems,val);
		FCount := Val;
	end else SetLength(FItems,val);
end;

function TSongDataArray.Add;
begin
	if FCount = GetCapacity then SetCapacity(GetListGrow(Capacity));

	Result := AddID(FCount);
	inc(FCount);
end;

procedure TSongDataArray.Delete(ID: Integer);
var
  List_CNT: Integer;
begin
	FItems[ID].Free;

	for List_CNT := ID to FCount - 2 do
		FItems[List_CNT] := FItems[List_CNT + 1];

	Dec(FCount);
end;

procedure TSongDataArray.Insert(ID: Integer; val: TSongData);
var
  List_CNT: Integer;
begin         
	if FCount = GetCapacity then SetCapacity(GetListGrow(Capacity));

	for List_CNT := FCount downto ID + 1 do
		FItems[List_CNT] := FItems[List_CNT - 1];

	if Val = NIL then AddID(ID) else AddID(ID).Assign(val);

	Inc(FCount);
end;

procedure TSongDataArray.Clear;
var LC : Integer;
begin
	for LC := 0 to FCount - 1 do
		FItems[LC].Free;

	SetLength(FItems,0);
	FCount := 0;
end;

procedure TSongDataArray.LoadFromStream(Stream: TStream);
var
	LC, Max: Integer;
begin
	Clear;
	Stream.Read(Max,SizeOf(FCount));
	for LC := 0 to Max - 1 do begin
		Self.Add();
		Self[LC].LoadFromStream(Stream);
	end;
end;

procedure TSongDataArray.LoadFromFile(const FileName: string);
var
	FStrings:TFileStream;
begin
	FStrings := TFileStream.Create(FileName,fmOpenRead);
	try
		LoadFromStream(FStrings);
	finally
		FStrings.Free;
	end;
end;

procedure TSongDataArray.SaveToStream(Stream: TStream);
var
	LC: Integer;
begin
	Stream.Write(FCount,SizeOf(FCount));
	for LC := 0 to FCount - 1 do begin
		Self[LC].SaveToStream(Stream);
	end;
end;

procedure TSongDataArray.SaveToFile(const FileName:String);
var
	FStrings : TFileStream;
begin
	FStrings := TFileStream.Create(FileName,fmCreate);
	SaveToStream(FStrings);
	FStrings.Free;
end;

procedure QuickSort(var FItems:Array of TSongData;L, R: Integer;Compare:TListSortCompare);
var
	I, J: Integer;
	P, T: TSongData;
begin
	repeat
		I := L;
		J := R;
		P := FItems[(L + R) shr 1];
		repeat
			while Compare(FItems[I], P) < 0 do Inc(I);
			while Compare(FItems[J], P) > 0 do Dec(J);
			if I <= J then begin
				T := FItems[I];
				FItems[I] := FItems[J];
				FItems[J] := T;
				Inc(I);
				Dec(J);
			end;
		until I > J;
		if L < J then QuickSort(FItems,L, J, Compare);
		L := I;
	until I >= R;
end;

procedure TSongDataArray.Sort(Compare: TListSortCompare);
begin
	if (FItems <> nil) and (Count > 0) then
		QuickSort(FItems,0, Count - 1,Compare);
end;

{ TSongDataLinker }

function TSongDataLinker.Add(Item: TSongData): Integer;
begin
	if FCount = GetCapacity then SetCapacity(GetListGrow(Capacity));

	FItems[FCount] := Item;
	Result := FCount;
	inc(FCount);
end;                     

procedure TSongDataLinker.AssignCommonData(Dest: TPersistent);
begin

end;

procedure TSongDataLinker.AssignTo(Dest: TPersistent);
var List_CNT: Integer;
begin
	if Dest is TSongDataArray then begin
		TSongDataLinker(Dest).Clear;
		TSongDataLinker(Dest).Capacity := GetCapacity;
		AssignCommonData(Dest);

		for List_CNT := 0 to FCount - 1 do begin
			TSongDataLinker(Dest).Add(FItems[List_CNT]);
		end;
	end else inherited;
end;

procedure TSongDataLinker.Clear;
begin
	SetCapacity(0);
	FCount := 0;
end;

procedure TSongDataLinker.Delete(ID: Integer);
begin

	if ID < FCount then
		System.Move(FItems[ID + 1], FItems[ID],
			(FCount - ID) * SizeOf(Pointer));

	{for List_CNT := ID to FCount - 2 do
		FItems[List_CNT] := FItems[List_CNT + 1];
  }
	Dec(FCount);
end;

destructor TSongDataLinker.Destroy;
begin
	Clear;
	inherited;
end;

function TSongDataLinker.GetCapacity: Integer;
begin
	if Assigned(FItems) then Result := Length(FItems)
	else Result := 0;
end;

function TSongDataLinker.GetItems(ID: Integer): TSongData;
begin
	Result := FItems[ID];
end;

function TSongDataLinker.IndexOf(Item: TSongData): Integer;
var LC: Integer;
begin
	for LC := 0 to Count - 1 do begin
		if GetItems(LC) = Item then begin
			Result := LC;
			Exit;
		end;
	end;
	Result := - 1;
end;

procedure TSongDataLinker.Insert(ID: Integer; val: TSongData);
begin
	if FCount = GetCapacity then SetCapacity(GetListGrow(Capacity));

	System.Move(FItems[ID],FItems[ID + 1],(FCount - ID) * SizeOf(Pointer));
	FItems[ID] := Val;
	Inc(FCount);
end;

procedure TSongDataLinker.LoadFromFile(const FileName: String);
var
	FStrings:TFileStream;
begin
	FStrings := TFileStream.Create(FileName,fmOpenRead);
	try
		LoadFromStream(FStrings);
	finally
		FStrings.Free;
	end;
end;

procedure TSongDataLinker.LoadFromStream(Stream: TStream);
var Max: Integer;
begin
	Clear;
	Stream.Read(Max,SizeOf(FCount));
	SetCapacity(Max);
	FCount := Max;
end;

procedure TSongDataLinker.SaveToFile(const FileName: String);
var
	FStrings : TFileStream;
begin
	FStrings := TFileStream.Create(FileName,fmCreate);
	SaveToStream(FStrings);
	FStrings.Free;
end;

procedure TSongDataLinker.SaveToStream(Stream: TStream);
begin
	Stream.Write(FCount,SizeOf(FCount));
end;

procedure TSongDataLinker.SetCapacity(const val: Integer);
begin
	if Val < GetCapacity then begin
		SetLength(FItems,val);
		FCount := Val;
	end else SetLength(FItems,val);
end;

procedure TSongDataLinker.SetItems(ID: Integer; val: TSongData);
begin
	FItems[ID] := Val;
end;

procedure TSongDataLinker.Sort(Compare: TListSortCompare);
begin
	if (FItems <> nil) and (Count > 0) then
		QuickSort(FItems,0, Count - 1,Compare);
end;

{TSongDataList Funx.}
Destructor TSongDataList.Destroy;
begin
	Clear;
	inherited;
end;             

function TSongDataList.GetCount:Integer;
begin
	Result := 0;
	if Assigned(FItems) then begin
		Result :=Length(FItems);
	end;
end;

procedure TSongDataList.AssignTo(Dest: TPersistent);
var
	LC : Integer;
begin
	if Dest is TSongDataList then begin
		TSongDataList(Dest).Clear;
		SetLength(TSongDataList(Dest).FItems,GetCount);
		for LC := 0 to GetCount - 1 do begin
			if not Assigned(FItems[LC]) then Continue;
			TSongDataList(Dest).AddID(LC).Assign(FItems[LC]);
		end;
	end;
end;

procedure TSongDataList.SetItems(ID: Integer; const val: TSongData);
begin
	FItems[ID].Assign(val);
end;

function TSongDataList.GetItems(ID: Integer) :TSongData;
begin
	Result := FItems[ID];
end;

function TSongDataList.Add(out ID: Integer):TSongData;
var
	List_CNT: Integer;
begin
	ID := - 1;

	for List_CNT := 0 to GetCount - 1 do begin
		if not Assigned(FItems[List_CNT]) then begin
			ID := List_CNT;
			break;
		end;
	end;

	if ID = - 1 then begin
		SetLength(FItems,GetCount + 1);
		ID := GetCount - 1;
	end;

	Result := AddID(ID);
end;

function TSongDataList.Add():TSongData;
var
	ID:Integer;
begin
	Result := Add(ID);
end;

procedure TSongDataList.Delete(ID: Integer);
begin
	FItems[ID].Free;
	FItems[ID] := NIL;
end;

procedure TSongDataList.Clear;
var
	Pilot_CNT: Integer;
begin
	for Pilot_CNT := 0 to Count - 1 do
		FItems[Pilot_CNT].Free;

	SetLength(FItems,0);
end;

procedure TSongDataList.Sort(Compare: TListSortCompare);
	procedure QuickSort(L, R: Integer);
	var
		I, J: Integer;
		P, T: TSongData;
	begin
		repeat
			I := L;
			J := R;
			P := FItems[(L + R) shr 1];
			repeat
				while Compare(FItems[I], P) < 0 do Inc(I);
				while Compare(FItems[J], P) > 0 do Dec(J);
				if I <= J then begin
					T := FItems[I];
					FItems[I] := FItems[J];
					FItems[J] := T;
					Inc(I);
					Dec(J);
				end;
			until I > J;
			if L < J then QuickSort(L, J);
			L := I;
		until I >= R;
	end;
begin
  if (FItems <> nil) and (Count > 0) then
		QuickSort(0, Count - 1);
end;

procedure TSongDataList.LoadFromFile(const FileName: string);
var
	FStrings:TFileStream;
begin
	FStrings := TFileStream.Create(FileName,fmOpenRead);
	Clear;
	try
		LoadFromStream(FStrings);
	finally
		FStrings.Free;
	end;
end;

procedure TSongDataList.LoadFromStream(Stream: TStream);
var
	MCount : Word;
	LC: Integer;
	B : Byte;
begin
	Clear;
	Stream.Read(MCount,2);
	SetLength(FItems,MCount);
	for LC := 0 to MCount - 1 do begin
		Stream.Read(B,1);
		if B = 255 then Continue;
		Stream.Position := Stream.Position - 1;
		Self.Add();
		Self[LC].LoadFromStream(Stream);
	end;
end;

procedure TSongDataList.SaveToFile(const FileName:String);
var
	FStrings : TFileStream;
begin
	FStrings := TFileStream.Create(FileName,fmCreate);
	SaveToStream(FStrings);
	FStrings.Free;
end;

procedure TSongDataList.SaveToStream(Stream: TStream);
var
	MCount : Word;
	LC: Integer;
const
	B : Byte = 255;
begin
	MCount := GetCount;
	Stream.Write(MCount,2);
	for LC := 0 to MCount - 1 do begin
		if Self[LC] = NIL then begin
			Stream.Write(B,1);
		end else begin
			Self[LC].SaveToStream(Stream);
		end;
	end;
end;

end.

